/* There are several parts to this file

     Part I : Implementation of specific generic tlv commands:
               To - to_command , Func - func_command , Print - print_command,
               Let - assign_command,  Quit - exit_command.

     Part II : Script execution - run_stmts, run_1_stmt

     Part III : Running procedures and function: map_params,
                run_command, run_func

     Part IV : load_script, go_interactive, init_script

   Changes:
   11/17/2003: Ittai -
     Added dynamic default parameters to procedures/functions. "dynamic" means
     that a parameter's default value is an expression that is evaluated
     whenever the procedure/function is called (rather than when the procedure/
     function is defined).

   The function gnu_getcwd defined here is taken from the gnu libc documentation
 */

#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <node.h>
#include <y.tab.h>
#include <node_smv.h>
#include <storage.h>
#include <hash.h>
#include <assoc.h>
#include <stdlib.h>
#include <stdio.h>
#include <util_err.h>
#include <string.h>
#include <mystring.h>
#include <script.h>
#include <symbols.h>
#include <eval.h>

int silent;

/* Hash tables which contain definitions of user defined
   procedures and functions. */
static hash_ptr progs_hash;
static hash_ptr funcs_hash[2];

static char *current_load_path = NULL;

#define PATH_SEP '/'
#define BUFSIZE 1000
char buf[BUFSIZE];

/* Pointers to functions. tlv.c assigns into these pointers. */

void *(*script_eval_node)(node_ptr parm, node_ptr n,node_ptr context, int typ);
int (*script_eval_node_to_boolean)(node_ptr n, node_ptr context);
int (*script_eval_num)(node_ptr num,node_ptr context);
node_ptr (*script_eval_node_to_symbol_key)
           (node_ptr n, node_ptr context);


void *(*script_null_value)(int typ);
void (*script_free_value)(void *val);

void (*script_insert_dynamic_var)(node_ptr parm_key,void *val);

void (*script_eval_node_and_print)(node_ptr n);
void (*script_assign_command)(node_ptr lhs,
                               node_ptr rhs);
void (*script_dump_command)(node_ptr n,
                             node_ptr fname,
                             node_ptr b);
void (*script_declare_var)(node_ptr n);


int (*script_application_command)(node_ptr stmt,node_ptr args);
void *(*script_application_function[5])(node_ptr name,node_ptr args);


/* Lists of all defined procedures and functions. */
static node_ptr list_of_procs;
static node_ptr list_of_funcs;

/* Print list of nodes. */
void print_node_list(node_ptr l)
{
  for(;l;l=cdr(l)) {
      print_node(tlvstdout,car(l));
      fprintf(tlvstdout,"\n");
  }
}

/* Print all defined procedures. */
void print_procs(void)
{
  fprintf(tlvstdout,"Procedures:\n");

  print_node_list(list_of_procs);
}

/* Print all defined functions. */
void print_funcs(void)
{
  fprintf(tlvstdout,"Functions:\n");

  print_node_list(list_of_funcs);
}


/***********************************************************************/

/* Interactive command to define new programs */
void to_command(node_ptr n, node_ptr param, node_ptr stmts)
{
  /* Find canonical name of procedure. */
  node_ptr name = find_atom(n);
  node_ptr val;

  /* If the routine already exists then erase it */
  if (val = find_assoc(progs_hash, name))
    remove_assoc(progs_hash, name, val);
  else
    list_of_procs = cons(name, list_of_procs);

  /* Insert the routine into the program hash */
  insert_assoc(progs_hash, name, find_node(LIST, param, stmts));

  if (!silent)
    fprintf(tlvstderr, "The program %s has been recorded\n",
            name->left.strtype->text);
}


/* Interactive command to define new functions */
void func_command(node_ptr n, node_ptr param, node_ptr stmts, int typ)
{
  /* Find cannonical name of function. */
  node_ptr name = find_atom(n);
  node_ptr val;

  /* If the routine already exists then erase it */
  if(val=find_assoc(funcs_hash[typ],name))
    remove_assoc(funcs_hash[typ],name,val);
  else
    list_of_funcs = cons(name,list_of_funcs);

  /* Insert the routine into the program hash */
  insert_assoc(funcs_hash[typ],name,find_node(LIST,param,stmts));

  if (!silent)
    fprintf(tlvstderr,"The function %s has been recorded\n",
                   name->left.strtype->text);
 }

/* s is a string. Substitue the two letter string "\n" with
   a space followed by a real \n . This allows the user
   to use "\n" within the Print command, in order to terminate
   line. */
char *subs_newline(char *s)
{
  char *pos;

  for (pos = s; pos = strstr(pos, "\\"); pos++)
    {
      int find = 1;

      switch (pos[1]) {
        case 'n': pos[0]='\n'; break; //newline
        case 't': pos[0]='\t'; break; //tab
        case 'r': pos[0]='\r'; break; //carriage return
        case 'f': pos[0]='\f'; break; //form feed 
        case 'v': pos[0]='\v'; break; //vertical tab
        case 'b': pos[0]='\b'; break; //backspace
        case 'a': pos[0]='\a'; break; //bell
        default: find=0; break;

        //case '"': pos[0]='"'; break; //double quote
        //this one is making some problems since you are 
        //probably locking for this char first
      }

      if (find) {
        // forcing the cursor back to place, that is one character before
        strcpy(pos+1, pos+2);
      }
    }
  return s;
}

/* Implementation of Print command.
   The items are printed in reverse. */
void print_command(node_ptr n)
{
  if (n) {
    node_ptr curr_expr = car(n);

    /* Recursively print next item on the list. */
    print_command(cdr(n));

    if (curr_expr->type == QUOTE)
      /* The current expression is a string, so print it. */
      fprintf(tlvstdout, "%s", subs_newline(curr_expr->left.strtype->text));

    else if (curr_expr->type == DEFINE) {
      /* The parameter is preceeded by ' . This means
         that the parameter should not be evaluated as a bdd, but
         as a symbol/parse-tree */

      /* Evaluate parameter as a symbol. */
      node_ptr val = eval_symbol(car(curr_expr));

      /* Print result. */
      strcpy(buf, "");
      sprint_node(buf, BUFSIZE, val);
      fprintf(tlvstdout, "%s", buf);
    }
    else
      /* Use tlv.c specific routine to interpret parameter and 
         print it. */
      if (script_eval_node_and_print)
        (*script_eval_node_and_print)(curr_expr);
      else
        fprintf(tlvstdout,"**cant print data**");
  }
}


/* Implement assign and dump comamnds by calling tlv.c code. */
void assign_command(node_ptr lhs,node_ptr rhs)
{
  (*script_assign_command)(lhs,rhs);
}
void dump_command(node_ptr n, node_ptr fname, node_ptr b)
{
  (*script_dump_command)(n,fname,b);
}

/* Exit command. Exit from program. */
void exit_command(node_ptr num)
{
  if (num == NIL) exit(0);

  exit((*script_eval_num)(num,NIL));
}


/***********************************************************************
     Part II : Script execution - run_stmts, run_1_stmt
 ***********************************************************************/

/* Run a sequence (given as a list) of statements. */
/* The returned result already has reference counters to it. */
void *run_stmts(node_ptr stmts, int typ)
{
  node_ptr curr_node = stmts;
  node_ptr curr_stmt;

  node_ptr stmt_stack = NIL;

  void *result;

  /* This loop runs a statment list which may contain nested
     statement lists. The inner loop simply goes through
     a list of statements and executes them. However, when
     there are nested lists such as when there is an IF statement,
     then execution moves on to a new list (the list either for
     the THEN, or the ELSE), but when that list is done we
     have to go back to resume execution of the original list.
     This is done by having a stack of the locations where
     we need to resume execution. This stack is observed
     in the outer loop. */
  while (curr_node) {
    /* This loop runs a statement list. */
    while (curr_node) {
      /* Get the current statement from the list. */
      curr_stmt = curr_node->left.nodetype;

      /* Execute statement */
      switch (curr_stmt->type) {
      case PRINT: print_command(car(curr_stmt)); break;
      case SETTIME: settime_command();  break;
      case CHKTIME: chktime_command();  break;
      case WHILE:
        /* Evaluate the condition as a boolean. */
        if ((*script_eval_node_to_boolean)(curr_stmt->left.nodetype,NIL)) {

          /* Remember our current position in a stack */
          push_node(&stmt_stack,curr_node);

          /* Start excuting a new statement list */
          curr_node = curr_stmt->right.nodetype;

          /* Skip over code at end loop which advances to the next statement */
          continue;
        }
        break;

      case IF: {
        /* Evaluate the condition as a boolean. */
        int result =
          (*script_eval_node_to_boolean)(condition_of_if(curr_stmt), NIL);

        if (result || (!result && else_of_if(curr_stmt) != NIL)) {
          /* Remember the next statement to be executed after
             this IF structure */
          if (curr_node->right.nodetype)
            push_node(&stmt_stack,curr_node->right.nodetype);

          /* Start executing a new statement list */
          curr_node = result ? then_of_if(curr_stmt) : else_of_if(curr_stmt);

          /* Skip over the code which advances to the next statement */
          continue;
        }

        break;
      }
      case TLVCASE: {
        /* Case command. */
        node_ptr scan;
        node_ptr str_list;

        /* Evaluate the term in the parenthesis as a parse tree. */
        node_ptr result = eval_symbol(car(curr_stmt));

        strcpy(buf,"");
        sprint_node(buf,BUFSIZE,result);

        /* Scan the list of strings and find one that matches. */
        for (scan=cdr(curr_stmt); scan; scan=cdr(scan))
          if (scan->type == DEFAULT)
            break;
          else
            for (str_list = car(car(scan));str_list; 
                 str_list = cdr(str_list))
              if (!strcmp(buf, atom_to_string(car(str_list))))
                goto get_out;
        get_out:
        /* Execute the statement sequence (if string found). */
        if (scan) {
          /* Remember the next statement to be executed after
             this CASE structure */
          if (curr_node->right.nodetype)
            push_node(&stmt_stack,curr_node->right.nodetype);

          /* Start executing a new statement list */
          curr_node = cdr(car(scan));

          /* We do this "continue" to skip over the code
             at the end of this loop which advances to
             the next statement */
          continue;
        }

        /* If no matching string was found, then just continue
           execution after the CASE statement. */
        break;
      }

      case RUN:
        run_command(curr_stmt->left.nodetype,curr_stmt->right.nodetype);
        break;

      case BREAK: case CONTINUE:
        /* Search for last loop in stack. Pop stack until we reach it.  */
        do {
          curr_node = pop_node(&stmt_stack);
        } while (curr_node != NIL && ! is_loop_stmt(car(curr_node)) );
        /* curr_node->left.nodetype->type != WHILE ); */

        if (curr_stmt->type == BREAK)
          break;
        else
          continue; /* skip over advancing curr_node to next statement.
                       This will bring us back to the beginning of the loop. */
      case TLV_RETURN:
        /* Empty stack */
        while (curr_node = pop_node(&stmt_stack));

        /* Evaluate result. */
        if (curr_stmt->left.nodetype)
          result = (*script_eval_node)(NIL,curr_stmt->left.nodetype,NIL, typ);
        else
          result = (*script_null_value)(typ);

        return (result);

      case EXIT:
        exit_command(car(curr_stmt));
        break;

        /* The reason we keep the next three cases in here and not
           in "application_command" is that if we want to create a
           program which will be able to understand different languages
           and data structures according to the input, then grammar.y
           will have to call different functions for assignment (for
           example) if there is no "generic" script_assign_command.
           Instead we  have
           a generic "assign_command" implemented by "script_assign_command".
           the assign_command supplied by script.h provides a generic
           assign command which can be used from grammar.y */

      case LOCAL: {
        node_ptr name_key =
          (*script_eval_node_to_symbol_key)(curr_stmt->left.nodetype, NIL);

        if (!exists_local(name_key))
          /* If such a variable already exists, then store
             its value so we can restore it later. */
          delete_symbol_and_store_on_frame(name_key);
        /* else */
        /*   fprintf(tlvstdout, "LOCAL: Skipping re-declaration of local\n"); */

        if (curr_stmt->right.nodetype == NIL)
          break;
      }
        /* fall through on purpose.*/

      case LET:
        (*script_assign_command)(curr_stmt->left.nodetype,
                                 curr_stmt->right.nodetype);
        break;

      case DUMP:
        (*script_dump_command)(curr_stmt->left.nodetype,
                               curr_stmt->right.nodetype->left.nodetype,
                               curr_stmt->right.nodetype->right.nodetype);
        break;

      case COLON:
        /* Declare a new program variable. */
        (*script_declare_var)(curr_stmt);
        break;

      default:
        /* fprintf(tlvstderr, "run_stmts: Strange statement -- %d\n", */
        /*         curr_stmt->type); */
        break;
      }

      /* Go to the next statement in the list. */
      if (curr_node != NIL)
        curr_node = curr_node->right.nodetype;
    }

    /* Return from nested list. */
    curr_node = pop_node(&stmt_stack);
  }
  return (*script_null_value)(typ);
}

/* Run one statement (this is used for running While and IF from
   command line) */
void run_1_stmt(node_ptr stmt)
{
  /* Wrap statement in a list. */
  node_ptr stmts = new_node(LIST, stmt, NIL);

  void *val = run_stmts(stmts, 0);

  (*script_free_value)(val);

  /* Free node used to wrap statement in list. */
  free_node(stmts);
}

/***********************************************************************
     Part III : Running procedures and function: map_params,
                run_command, run_func
 ***********************************************************************/

/* eval_arg: Evaluates an argument corresponding to a given parameter */
node_ptr eval_arg(node_ptr param, node_ptr arg)
{
  node_ptr result;
  if (param->type == INOUT) {
    /* This handles passing parameters by name. This occurs when the
       formal paramter is preceded by &  . */

    /* I don't know which of these two lines is better. The additional
       eval_struct may provide easier calls with the same parameter name,
       since the argument passed to the next parameter would not be the parm
       name, but its value. */

    result = eval_struct(arg, NIL);
    /* result = new_node(LIST,arg,NIL); */
  }
  else if (param->type == DEFINE)
    /* Symbolic evaluation, when formal paramter is preceeded by ' */
    result = eval_symbol(arg);
  else {
    node_ptr parm_key = eval_struct(param, NIL); /* Find node for parameter */

    /* Evaluate argument. */
    result = (node_ptr) (*script_eval_node)(parm_key,arg,NIL,0);
  }
  return result;
}

/* Associate actual parameters with formal parameters. */
/* Loop over both parameter lists .                    */
/* Also pushes new frame, but only if succeded in performing the map,
   (which is indicated by a return value of 1. */
static int map_params(char *name, node_ptr arg_list, node_ptr parm_list)
{
  node_ptr curr_arg = arg_list;
  node_ptr curr_parm;

  /* Length of parameter lists. */
  int len_p = length(parm_list);
  int len_a = length(arg_list);

  int params_to_default = len_p - len_a;

  /* List of evaluated arguments. */
  node_ptr eval_arg_list = NIL;

  /* Check that the number of formal parameters is greater than the number
    of actual arguments. If there are leftover parameters then hopefully they
    have default values. Otherwise we'll report an error later on. */
  if (len_p < len_a) {
    fprintf(tlvstderr, "The number of parameters is incorrect\n");
    fprintf(tlvstderr,
            "The routine %s expects %d parameters but received %d\n", name,
            len_p, len_a);
    return 0;
  }
  else if (len_p > len_a) {
    /* There are more formal parameters than arguments.
       We need to check that there are at least len_p - len_a
       default parameters. */
    int i = params_to_default;

    for (curr_parm = parm_list;  curr_parm != NIL && i > 0; 
         curr_parm = cdr(curr_parm)) {

      node_ptr parm = car(curr_parm);
      i--;

      if (!is_default_param(parm)) {
        fprintf(tlvstderr, "The number of parameters is incorrect\n");
        fprintf(tlvstderr,
                "The routine %s expects at least %d parameters but received"
                " %d\n",
                name, len_a + 1, len_a);
        return 0;
      }
    }
  }

  /*** Evaluate arguments ***/

  /* Loop which evaluates arguments and adds them to the list
     "eval_arg_list" */
  for (curr_parm = parm_list ; curr_parm != NIL; curr_parm = cdr(curr_parm) ) {
    node_ptr parm = curr_parm->left.nodetype;
    node_ptr arg;

    if (params_to_default == 0) {
      arg = curr_arg->left.nodetype;
      /* Go to next argument */
      curr_arg = curr_arg->right.nodetype;
    }
    else {
      params_to_default--;
      if (is_default_param(parm))
        arg = default_param_value(parm);
      else
        catastrophe("Error: in map_params");
    }
    /* Unwrap a default param */
    if (is_default_param(parm))
      parm = default_param_formal(parm);

    eval_arg_list = append(eval_arg_list,
                           new_node(LIST, eval_arg(parm, arg), NIL));
  }

  /* We can push the frame only after evaluating the arguments since
     evaluating the arguments may require the previous frame. */
  push_frame();

  /*** Map evaluated arguments list. ***/

  /* Loop which maps actual parameters to formal paramters */
  curr_arg = eval_arg_list;
  for (curr_parm = parm_list; curr_parm != NIL; curr_parm = cdr(curr_parm)) {
    node_ptr parm = curr_parm->left.nodetype;
    node_ptr arg = curr_arg->left.nodetype;

    node_ptr parm_key;

    if (parm->type == EQDEF)
      parm = parm->left.nodetype;
    if (parm->type == INOUT) {
      parm_key = eval_struct(car(parm), NIL); /* Find node for parameter */
      insert_by_name_param(parm_key, arg);    /* Associate parameter by name. */
    }
    else if (parm->type == DEFINE) {
      parm_key = eval_struct(car(parm), NIL);  /* Find node for parameter */
      /* Remember previous definition. */
      delete_symbol_and_store_on_frame(parm_key);
      insert_parse_tree(parm_key, arg);       /* Create parse tree variable */
    }
    else {
      parm_key = eval_struct(parm, NIL);  /* Find node for parameter */

      /* Remember previous definition of symbol. */
      delete_symbol_and_store_on_frame(parm_key);

      /* Insert dyname var. */
      if (script_insert_dynamic_var)
        (*script_insert_dynamic_var)(parm_key, (void *) arg);
      else
        insert_dynamic_var(parm_key, (bdd_ptr) arg);
    }
    curr_arg = curr_arg->right.nodetype; /* Go to next argumenmt */
  }

  free_list(eval_arg_list);

  return 1;
}

/* Interactive command to run programs. name is the name of subprogram
   to run, and args are the arguments. */
void run_command(node_ptr name, node_ptr args)
{
  node_ptr prg;

  /* This is a list which contains the previous values of variables
     which were overrun by local variables.
     They are restrored after this routines ends it's execution.
     node_ptr global_list= NIL ;  */

  /* Find program in prog hash */
  node_ptr name_node = find_node(name->type,name->left.nodetype,
                                            name->right.nodetype);
  if (name_node == NIL ||
      (prg= find_assoc(progs_hash,name_node)) == NIL) {
    /* No user defined command found.
       Try to execute a predefined application command from tlv.c . */
    if (script_application_command &&
        (*script_application_command)(name_node, args))
      return;
    else {
      fprintf(tlvstderr, "The program %s is not defined\n",
              name->left.strtype->text);
      return;
    }
  }

  /* Map arguments to formal parameters, and execute routine.
     push frame is done inside map_params. */
  if (map_params(name->left.strtype->text,
                 args,prg->left.nodetype)) {
    void *val = run_stmts(prg->right.nodetype, 0); /* Run program */
    (*script_free_value)(val);
    pop_frame();
  }
}

/* Interactive command to run programs */
void *run_func(node_ptr name, node_ptr args, int typ)
{
  void *result;
  node_ptr name_node = find_atom(name);
  node_ptr prg;

  /* Try to find a user defined function. */
  if ((prg = find_assoc(funcs_hash[typ],name_node)) == NIL) {
    /* No user defined function found.
       Try to execute a predefined application function. */
    if (script_application_function[typ])
      if ((result = (*script_application_function[typ])(name_node,args)) !=
          NULL)
        return result;

    /* Only print error for a function returning a bdd. */
    if (typ == 0)
      fprintf(tlvstderr, "The function %s is not defined\n",
              name->left.strtype->text);
    return (*script_null_value)(typ);
  }

  /* Execute user defined function. */

  /* push frame is done inside map_params. */
  if (map_params(name->left.strtype->text,
                 args,prg->left.nodetype)) {
    /* Run the program */
    result = run_stmts(prg->right.nodetype, typ);
    pop_frame();

    return result;
  }
  else
    return (*script_null_value)(typ);
}

/***********************************************************************/

/// Taken from gnu libc's info documentation
char *gnu_getcwd()
{
  size_t size = 100;
  while (1) {
    char *buffer = (char *) malloc(sizeof(char) * size);
    if (getcwd(buffer, size) == buffer)
      return buffer;
    free(buffer);
    if (errno != ERANGE)
      return NULL;
    size *= 2;
  }
}

//Create a new array.
void load_new_arr()
{
	node_ptr name = string_to_atom("new");
	node_ptr atom_arr = string_to_atom("arr");
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr param = cons(new_node(INOUT, atom_arr, NIL), NIL);
	node_ptr myarr_0 = new_node(ARRAY, atom_arr, num_0);
	node_ptr ifcond = new_node(EXIST, myarr_0, NIL); 
	node_ptr ifact = cons(cons(new_node(RUN, string_to_atom("delarr"), 
					cons(atom_arr, NIL)), NIL), NIL);
	node_ptr nxtact = cons(new_node(LET, myarr_0, num_0), NIL);
	node_ptr stmts = cons(new_node(IF, ifcond, ifact), nxtact);
	to_command(name, param, stmts);
}

//Delete an array.
void load_del_arr()
{
	node_ptr name = string_to_atom("delarr");
	node_ptr atom_arr = string_to_atom("arr");
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_i = string_to_atom("i");
	node_ptr myarr_0 = new_node(ARRAY, atom_arr, num_0);
	node_ptr myarr_i = new_node(ARRAY, atom_arr, num_i);

	node_ptr param = cons(new_node(INOUT, atom_arr, NIL), NIL);

	node_ptr foract  = cons(cons(new_node(RUN, string_to_atom("delvar"), 
					cons(myarr_i, NIL)), NIL), NIL);
	node_ptr nxtact  = cons(cons(new_node(RUN, string_to_atom("delvar"), 
					cons(myarr_0, NIL)), NIL), NIL);
	node_ptr stmts = cons(new_tlvfor(num_i, PLUS, num_1, myarr_0, reverse(foract)), nxtact);
	to_command(name, param, stmts);
}

//Push arr the val.
void load_push_arr()
{
	node_ptr name = string_to_atom("push");
	node_ptr atom_arr = string_to_atom("arr");
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr myarr_0 = new_node(ARRAY, atom_arr, num_0);

	node_ptr param = cons(string_to_atom("val"), cons(new_node(INOUT, atom_arr, NIL), NIL));

	node_ptr ifcond = new_node(NOT, new_node(EXIST, myarr_0, NIL), NIL); 
	node_ptr ifact = cons(new_node(LET, myarr_0, num_0), NIL);
	node_ptr ifstmt = new_node(IF, ifcond, ifact);
	node_ptr nxtact1 = new_node(LET, myarr_0, new_node(PLUS, myarr_0, num_1));
	node_ptr nxtact2 = new_node(LET, new_node(ARRAY, atom_arr, myarr_0), string_to_atom("val"));

	node_ptr stmts = cons(nxtact2, cons(nxtact1, cons(ifstmt, NIL)));
	to_command(name, param, reverse(stmts));
}

//Pop the top element of arr
void load_pop_arr()
{
	node_ptr name = string_to_atom("pop");
	node_ptr atom_arr = string_to_atom("arr");
	node_ptr atom_top = string_to_atom("top");
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_m1 = new_NUMBER_node(-1);
	node_ptr myarr_0 = new_node(ARRAY, atom_arr, num_0);
	node_ptr atom_res = string_to_atom("result");

	node_ptr param = cons(new_node(INOUT, atom_arr, NIL), NIL);
	node_ptr ifcond = myarr_0;

	node_ptr localact = new_node(LOCAL, atom_res, new_node(FUNC, atom_top, 
				cons(atom_arr, NIL)));
	node_ptr delvaract  = new_node(RUN, string_to_atom("delvar"), 
					cons(new_node(ARRAY, atom_arr, myarr_0), NIL));
	node_ptr letact = new_node(LET, myarr_0, new_node(MINUS, myarr_0, num_1));
	node_ptr retact = new_node(TLV_RETURN, atom_res, NIL);
	node_ptr thenact = cons(retact, cons(letact, cons(delvaract, cons(localact, NIL))));

	node_ptr prtact = new_node(PRINT, cons(string_to_atom("Warning: stack is empty\n"), NIL), NIL);
	node_ptr retact1 = new_node(TLV_RETURN, num_m1, NIL);
	node_ptr elseact = cons(retact1, cons(prtact, NIL));
	
	node_ptr ifstmt = new_node(IF, ifcond, cons(reverse(thenact),reverse(elseact)));

	node_ptr stmts = cons(ifstmt, NIL);
	func_command(name, param, reverse(stmts), 0);
}

//The top element of arr
void load_top_arr()
{

	node_ptr name = string_to_atom("top");
	node_ptr atom_arr = string_to_atom("arr");
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr myarr_0 = new_node(ARRAY, atom_arr, num_0);

	node_ptr param = cons(new_node(INOUT, atom_arr, NIL), NIL);
	node_ptr retact = new_node(TLV_RETURN, new_node(ARRAY, atom_arr, myarr_0), NIL);
	node_ptr stmts = cons(retact, NIL);
	func_command(name, param, reverse(stmts), 0);
}

//The length of arr
void load_length_arr()
{

	node_ptr name = string_to_atom("length");
	node_ptr atom_arr = string_to_atom("arr");
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr myarr_0 = new_node(ARRAY, atom_arr, num_0);

	node_ptr param = cons(new_node(INOUT, atom_arr, NIL), NIL);
	node_ptr retact = new_node(TLV_RETURN, myarr_0, NIL);
	node_ptr stmts = cons(retact, NIL);
	func_command(name, param, reverse(stmts), 0);
}

void load_new_ts()
{
	node_ptr name = string_to_atom("new_ts");
	node_ptr a_type = string_to_atom("type");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_s_top = new_node(DOT, a_s, string_to_atom("top"));
	node_ptr arr_s_top = new_node(ARRAY, a_s, a_s_top);
	node_ptr a_true = new_node(TRUEEXP, NIL, NIL);
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_2 = new_NUMBER_node(2);

	node_ptr param = cons(new_default_param(a_type, num_2), NIL);

	node_ptr act1 = new_node(LET, a_s_top, new_node(PLUS, a_s_top, num_1)); 
	node_ptr act2 = new_node(LET, new_node(DOT, arr_s_top, a_type), a_type);
	node_ptr act3 = new_node(LET, new_node(DOT, arr_s_top, string_to_atom("tn")), num_0);
	node_ptr act4 = new_node(LET, new_node(DOT, arr_s_top, string_to_atom("jn")), num_0);
	node_ptr act5 = new_node(LET, new_node(DOT, arr_s_top, string_to_atom("cn")), num_0);

	node_ptr act6 = new_node(LET, new_node(DOT, arr_s_top, string_to_atom("i")), a_true);
	node_ptr act7 = new_node(LET, new_node(DOT, arr_s_top, string_to_atom("v")), a_true);
	node_ptr retact = new_node(TLV_RETURN, a_s_top, NIL);

	node_ptr stmts = cons(act1, cons(act2, cons(act3, cons(act4, cons(act5, cons(act6, cons(act7, cons(retact, NIL)))))))); 
	func_command(name, param, stmts, 0);
}
//Initial condition
void load_set_I()
{
	node_ptr name = string_to_atom("set_I");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_i = string_to_atom("_i");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));
	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_i, NIL)); 
	node_ptr letact = new_node(LET, new_node(DOT, new_node(ARRAY, a_s, a_tsn), string_to_atom("i")), a_i);
	node_ptr stmts = cons(letact, NIL);
	to_command(name, param, stmts);
}

void load_I()
{
	node_ptr name = string_to_atom("I");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));
	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), NIL); 
	node_ptr retact = new_node(TLV_RETURN, new_node(DOT, new_node(ARRAY, a_s, a_tsn), string_to_atom("i")), NIL);
	node_ptr stmts = cons(retact, NIL);
	func_command(name, param, stmts, 0);
}

//Variables
void load_set_V()
{
	node_ptr name = string_to_atom("set_V");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_v = string_to_atom("v");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));
	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_v, NIL)); 
	node_ptr letact = new_node(LET, new_node(DOT, new_node(ARRAY, a_s, a_tsn), a_v), a_v);
	node_ptr stmts = cons(letact, NIL);
	to_command(name, param, stmts);
}

void load_V()
{
	node_ptr name = string_to_atom("V");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));
	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), NIL); 
	node_ptr retact = new_node(TLV_RETURN, new_node(DOT, new_node(ARRAY, a_s, a_tsn), string_to_atom("v")), NIL);
	node_ptr stmts = cons(retact, NIL);
	func_command(name, param, stmts, 0);
}

//Add disjuncts to TS.
void load_add_disjunct_T()
{
	node_ptr name = string_to_atom("add_disjunct_T");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_i = string_to_atom("i");
	node_ptr a_t = string_to_atom("_t");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_t, NIL));

	node_ptr a_type = string_to_atom("type");
	node_ptr arr_s_tsn = new_node(ARRAY, a_s, a_tsn);
	node_ptr a_s_tsn_tn = new_node(DOT, arr_s_tsn, string_to_atom("tn"));
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_2 = new_NUMBER_node(2);

	node_ptr ifcond = new_node(EQUAL, new_node(DOT, arr_s_tsn, a_type), num_2);

	node_ptr act1 = new_node(LET, a_s_tsn_tn, new_node(PLUS, a_s_tsn_tn, num_1)); 
	node_ptr act2 = new_node(LOCAL, a_i, a_s_tsn_tn);
	node_ptr act3 = new_node(LET, new_node(ARRAY, new_node(DOT, arr_s_tsn, string_to_atom("t")), a_i), a_t);
	node_ptr thenact = cons(act1, cons(act2, cons(act3, NIL)));
	node_ptr ifstmt = new_node(IF, ifcond, cons(thenact, NIL));

	node_ptr stmts = cons(ifstmt, NIL);
	to_command(name, param, stmts);
}

//Return i-th component of transistion system.
void load_Ti()
{
	node_ptr name = string_to_atom("Ti");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_i = string_to_atom("i");
	node_ptr a_t = string_to_atom("t");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_i, NIL));
	//node_ptr param = cons(a_i, cons(new_default_param(a_tsn, a_s_cur), NIL));

	node_ptr a_type = string_to_atom("type");
	node_ptr arr_s_tsn = new_node(ARRAY, a_s, a_tsn);
	node_ptr a_s_tsn_t = new_node(DOT, arr_s_tsn, string_to_atom("t"));
	node_ptr a_s_tsn_t_comb = new_node(DOT, arr_s_tsn, string_to_atom("t_comb"));
	node_ptr a_s_tsn_t_pres = new_node(DOT, arr_s_tsn, string_to_atom("t_pres"));
	node_ptr num_1 = new_NUMBER_node(1);

	node_ptr ifcond = new_node(EQUAL, new_node(DOT, arr_s_tsn, a_type), num_1);
	node_ptr thenact = cons(new_node(TLV_RETURN, new_node(AND, new_node(AND, new_node(ARRAY, a_s_tsn_t, a_i), new_node(ARRAY, a_s_tsn_t_comb, a_i)), new_node(ARRAY, a_s_tsn_t_pres, a_i)), NIL), NIL);

	node_ptr elseact = cons(new_node(TLV_RETURN, new_node(ARRAY, a_s_tsn_t, a_i), NIL), NIL); 
	node_ptr ifstmt = new_node(IF, ifcond, cons(thenact, elseact));
	node_ptr stmts = cons(ifstmt, NIL);
	func_command(name, param, stmts, 0);

}

//Find predecessors of transistion systems.
void load_pred1()
{
	node_ptr name = string_to_atom("pred1");
	node_ptr a_s = string_to_atom("_s");
	node_ptr as = string_to_atom("s");
	node_ptr a_i = string_to_atom("i");
	node_ptr a_tn = string_to_atom("tn");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_res = string_to_atom("result");
	node_ptr a_pred = string_to_atom("pred");
	node_ptr a_Ti = string_to_atom("Ti");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr arr_s_tsn = new_node(ARRAY, a_s, a_tsn);
	node_ptr a_s_tsn_tn = new_node(DOT, arr_s_tsn, string_to_atom("tn"));

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(as, NIL));

	node_ptr localact = new_node(LOCAL, a_res, num_0); 

	node_ptr tifunc = new_node(FUNC, a_Ti, cons(a_tsn, cons(a_i, NIL)));
	node_ptr letact = new_node(LET, a_res, new_node(OR, a_res, new_node(FUNC, a_pred, cons(as, cons(tifunc, NIL))))); 
	node_ptr foract = new_tlvfor(a_i, PLUS, num_1, a_s_tsn_tn, cons(letact, NIL));
	node_ptr retact = new_node(TLV_RETURN, a_res, NIL);
	//node_ptr stmts = cons(localact, cons(foract, cons(retact, NIL)));
	node_ptr stmts = cons(retact, append(foract, cons(localact, NIL)));

	func_command(name, param, reverse(stmts), 0);
}

//Find successors of transistion systems.
void load_succ1()
{
	node_ptr name = string_to_atom("succ1");
	node_ptr a_s = string_to_atom("_s");
	node_ptr as = string_to_atom("s");
	node_ptr a_i = string_to_atom("i");
	node_ptr a_tn = string_to_atom("tn");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_res = string_to_atom("result");
	node_ptr a_succ = string_to_atom("succ");
	node_ptr a_Ti = string_to_atom("Ti");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr arr_s_tsn = new_node(ARRAY, a_s, a_tsn);
	node_ptr a_s_tsn_tn = new_node(DOT, arr_s_tsn, string_to_atom("tn"));

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(as, NIL));

	node_ptr localact = new_node(LOCAL, a_res, num_0); 

	node_ptr tifunc = new_node(FUNC, a_Ti, cons(a_tsn, cons(a_i, NIL)));
	node_ptr letact = new_node(LET, a_res, new_node(OR, a_res, new_node(FUNC, a_succ, cons(as, cons(tifunc, NIL))))); 
	node_ptr foract = new_tlvfor(a_i, PLUS, num_1, a_s_tsn_tn, cons(letact, NIL));
	node_ptr retact = new_node(TLV_RETURN, a_res, NIL);
	//node_ptr stmts = cons(localact, cons(foract, cons(retact, NIL)));
	node_ptr stmts = cons(retact, append(foract, cons(localact, NIL)));

	func_command(name, param, reverse(stmts), 0);
}

//Find successors of transistion systems.
void load_successors()
{
	node_ptr name = string_to_atom("successors");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_set = string_to_atom("set");
	node_ptr a_cls = string_to_atom("closure");
	node_ptr a_frt = string_to_atom("front");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_succ = string_to_atom("succ1");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_set, NIL));

	node_ptr lcls = new_node(LOCAL, a_cls, a_set); 
	node_ptr lfrt = new_node(LOCAL, a_frt, a_set); 

	node_ptr letfrt = new_node(LET, a_frt, new_node(AND, new_node(FUNC, a_succ, cons(a_tsn, cons(a_frt, NIL))), new_node(NOT, a_cls, NIL))); 
	node_ptr letcls = new_node(LET, a_cls, new_node(OR, a_cls, a_frt));
	node_ptr fixact = new_fix(a_cls, cons(letfrt, cons(letcls, NIL)));

	node_ptr retact = new_node(TLV_RETURN, a_cls, NIL);
	node_ptr stmts = cons(retact, append(fixact, cons(lfrt, cons(lcls, NIL))));

	func_command(name, param, reverse(stmts), 0);
}


//Find predecessors of transistion systems.
void load_predecessors()
{
	node_ptr name = string_to_atom("predecessors");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_set = string_to_atom("set");
	node_ptr a_cls = string_to_atom("closure");
	node_ptr a_frt = string_to_atom("front");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_pred = string_to_atom("pred1");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_set, NIL));

	node_ptr lcls = new_node(LOCAL, a_cls, a_set); 
	node_ptr lfrt = new_node(LOCAL, a_frt, a_set); 

	node_ptr letfrt = new_node(LET, a_frt, new_node(AND, new_node(FUNC, a_pred, cons(a_tsn, cons(a_frt, NIL))), new_node(NOT, a_cls, NIL))); 
	node_ptr letcls = new_node(LET, a_cls, new_node(OR, a_cls, a_frt));
	node_ptr fixact = new_fix(a_cls, cons(letfrt, cons(letcls, NIL)));

	node_ptr retact = new_node(TLV_RETURN, a_cls, NIL);
	node_ptr stmts = cons(retact, append(fixact, cons(lfrt, cons(lcls, NIL))));

	func_command(name, param, reverse(stmts), 0);
}

//Computing the reachable states of transistion systems.
void load_reachable()
{
	node_ptr name = string_to_atom("reachable");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_I = string_to_atom("I");
	node_ptr a_succ = string_to_atom("successors");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), NIL);


	node_ptr fI = new_node(FUNC, a_I, cons(a_tsn, NIL));

	node_ptr retact = new_node(TLV_RETURN, new_node(FUNC, a_succ, cons(a_tsn, cons(fI, NIL))), NIL);
	node_ptr stmts = cons(retact, NIL);

	func_command(name, param, reverse(stmts), 0);
}

//Converting dynamic variables to work
void load_convert_old_tlv_to_TS()
{
	node_ptr name = string_to_atom("convert_old_tlv_to_TS");
	node_ptr a_t = string_to_atom("_t");
	node_ptr a_d = string_to_atom("_d");
	node_ptr a_p = string_to_atom("_pres");
	node_ptr ai = string_to_atom("i");
	node_ptr a_i = string_to_atom("_i");
	node_ptr a_si = string_to_atom("si");
	node_ptr a_sn = string_to_atom("_sn");
	node_ptr a_tn = string_to_atom("_tn");
	node_ptr a_type = string_to_atom("type");
	node_ptr a_newts = string_to_atom("new_ts");
	node_ptr a_add_disjunct_T = string_to_atom("add_disjunct_T");
	node_ptr a_set_I = string_to_atom("set_I");
	node_ptr a_set_V = string_to_atom("set_V");
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_0 = new_NUMBER_node(0);
	node_ptr a_vars = string_to_atom("_vars");
	node_ptr a_ct = string_to_atom("_curr_t");
	node_ptr a_cj = string_to_atom("_curr_just");
	node_ptr a_cc = string_to_atom("_curr_comp");

	node_ptr param = cons(a_type, NIL);

	node_ptr lct = new_node(LOCAL, a_ct, num_0);
	node_ptr lcj = new_node(LOCAL, a_cj, num_0);
	node_ptr lcc = new_node(LOCAL, a_cc, num_0);
	node_ptr lnewts = new_node(LOCAL, a_newts, NIL);
	node_ptr localact = cons(lnewts, cons(lcc, cons(lcj, cons(lct, NIL))));

	node_ptr arr_t = new_node(ARRAY, a_t, new_node(PLUS, a_ct, ai));
	node_ptr arr_d = new_node(ARRAY, a_d, new_node(PLUS, a_ct, ai));
	node_ptr arr_p = new_node(ARRAY, a_p, new_node(PLUS, a_ct, ai));
	node_ptr a_tdp = new_node(AND, new_node(AND, arr_t, arr_d), arr_p);
	node_ptr adT = new_node(RUN, a_add_disjunct_T, cons(a_newts, cons(a_tdp, NIL))); 
	node_ptr infor = new_tlvfor(ai, PLUS, num_1, new_node(ARRAY, a_tn, a_si), cons(adT, NIL));

	node_ptr nts = new_node(LET, a_newts, new_node(FUNC, a_newts, cons(a_type, NIL)));
	node_ptr seti = new_node(RUN, a_set_I, cons(a_newts, cons(new_node(ARRAY, a_i, a_si), NIL)));
	node_ptr setv = new_node(RUN, a_set_V, cons(a_newts, cons(new_node(ARRAY, a_vars, a_si), NIL)));
	node_ptr inact = cons(setv, cons(seti, cons(nts, NIL)));
	
	node_ptr inforact = append(infor, inact);
	node_ptr outfor = new_tlvfor(a_si, PLUS, num_1, a_sn, reverse(inforact));

	node_ptr stmts = append(outfor, localact);

	to_command(name, param, reverse(stmts));
}

//Model checking safty, using frontier.
void load_Cag2()
{
	node_ptr name = string_to_atom("Cag2");

	node_ptr a_s = string_to_atom("_s");
	node_ptr a_i = string_to_atom("i");
	node_ptr a_bulk = string_to_atom("bulk");
	node_ptr a_new = string_to_atom("new");
	node_ptr a_bpath = string_to_atom("bpath");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_ac = string_to_atom("arg_ceg");
	node_ptr a_pred = string_to_atom("pred1");
	node_ptr a_succ = string_to_atom("succ1");
	node_ptr a_len = string_to_atom("length");
	node_ptr a_delarr = string_to_atom("delarr");
	node_ptr a_ce = string_to_atom("ce");
	node_ptr a_push = string_to_atom("push");
	node_ptr a_pop = string_to_atom("pop");
	node_ptr a_top = string_to_atom("top");
	node_ptr a_fsat = string_to_atom("fsat");
	node_ptr a_I = string_to_atom("I");
	node_ptr a_V = string_to_atom("V");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_0 = new_NUMBER_node(0);

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_ac, NIL));

	node_ptr let_bulk1 = new_node(LET, a_bulk, new_node(AND, new_node(NOT, a_ac, NIL), new_node(FUNC, a_pred, cons(a_tsn, cons(num_1, NIL)))));	

	node_ptr loc_new = new_node(LOCAL, a_new, a_bulk);
	node_ptr prt_mcip = new_node(PRINT, cons(new_string_node(QUOTE, "Model checking Invariance Property\n", NIL), NIL), NIL);
	node_ptr new_bpath = new_node(RUN, a_new, cons(a_bpath, NIL));
	node_ptr push_b_b = new_node(RUN, a_push, cons(a_bulk, cons(a_bpath, NIL)));
//While(new)	
	node_ptr let_new1 = new_node(LET, a_new, new_node(AND, new_node(FUNC, a_pred, cons(a_tsn, cons(a_new, NIL))), new_node(NOT, a_bulk, NIL)));
	node_ptr push_b_n = new_node(RUN, a_push, cons(a_new, cons(a_bpath, NIL)));
	node_ptr let_bulk2 = new_node(LET, a_bulk, new_node(OR, a_bulk, a_new));
	node_ptr ifcond1 = new_node(AND, a_new, new_node(FUNC, a_I, cons(a_tsn, NIL)));
	node_ptr then1 = new_node(LET, a_new, num_0);
	node_ptr ifact1 = new_tlv_if(ifcond1, then1, NIL);
	
	node_ptr wstmts = cons(let_new1, cons(push_b_b, cons(let_bulk2, cons(ifact1, NIL))));
	node_ptr wact1 = new_node(WHILE, a_new, wstmts);
//End while(new)
//If(bulk &I(tsn))	
	node_ptr ifcond2 = new_node(AND, a_bulk, new_node(FUNC, a_I, cons(a_tsn, NIL)));
	//Then 
	node_ptr prt_nv = new_node(PRINT, cons(new_string_node(QUOTE, "\n*** Property is NOT VALID ***\n", NIL), NIL), NIL);
	node_ptr prt_ce = new_node(PRINT, cons(new_string_node(QUOTE, "\nCounter-Example Follows:\n", NIL), NIL), NIL);
	node_ptr new_ce = new_node(RUN, a_new, cons(a_ce, NIL));
	node_ptr afsat1 = new_node(FUNC, a_fsat, cons(new_node(FUNC, a_V, cons(a_tsn, NIL)), cons(new_node(AND, new_node(FUNC, a_pop, cons(a_bpath, NIL)), new_node(FUNC, a_I, cons(a_tsn, NIL))), NIL))); 
	node_ptr push_ce1 = new_node(RUN, a_push, cons(afsat1, cons(a_ce, NIL)));
		//While (length(ce))
	node_ptr afsat2 = new_node(FUNC, a_fsat, cons(new_node(FUNC, a_V, cons(a_tsn, NIL)), cons(new_node(AND, new_node(FUNC, a_pop, cons(a_bpath, NIL)), new_node(FUNC, a_succ, cons(a_tsn, cons(new_node(FUNC, a_top, cons(a_ce, NIL)), NIL)))), NIL))); 

	node_ptr push_ce2 = new_node(RUN, a_push, cons(afsat2, cons(a_ce, NIL)));
	node_ptr wact2 = new_node(WHILE, new_node(FUNC, a_len, cons(a_bpath, NIL)), cons(push_ce2, NIL));
		//End While (length(ce))
		//For
	node_ptr prt_st = new_node(PRINT, cons(new_node(ARRAY, a_ce, a_i), cons(new_string_node(QUOTE, "=\n", NIL), cons(a_i, cons(new_string_node(QUOTE, "\n--- State no. ", NIL), NIL)))), NIL);
	node_ptr foract = new_tlvfor(a_i, PLUS, num_1, new_node(FUNC, a_len, cons(a_ce, NIL)), cons(prt_st, NIL));
		//End for
	node_ptr ret0 = new_node(TLV_RETURN, num_0, NIL);
	//Else
	node_ptr preact = cons(wact2, cons(push_ce1, cons(new_ce, cons(prt_ce, cons(prt_nv, NIL))))); 
	node_ptr thenact2 = cons(ret0, append(foract, preact));
	node_ptr prt_iv = new_node(PRINT, cons(new_string_node(QUOTE, "\n*** Property is VALID ***\n", NIL), NIL), NIL);
	node_ptr delb = new_node(RUN, a_delarr, cons(a_bpath, NIL));
	node_ptr prt_bulk = new_node(PRINT, a_bulk, NIL);
	node_ptr ret1 = new_node(TLV_RETURN, num_1, NIL);
	node_ptr elseact2 = cons(ret1, cons(delb, cons(prt_iv, NIL)));
	//End Else
	node_ptr ifact2 = new_tlv_if(ifcond2, reverse(thenact2), reverse(elseact2));
//End If

	node_ptr stmts = cons(ifact2, cons(wact1, cons(push_b_b, cons(new_bpath, cons(prt_mcip, cons(loc_new, cons(let_bulk1, NIL)))))));

	func_command(name, param, reverse(stmts), 0);
}

//Model checking safty, using frontier.
void load_Cag()
{
	node_ptr name = string_to_atom("Cag");

	node_ptr a_s = string_to_atom("_s");
	node_ptr a_i = string_to_atom("i");
	node_ptr a_bulk = string_to_atom("bulk");
	node_ptr a_new = string_to_atom("new");
	node_ptr a_newf = string_to_atom("newf");
	node_ptr a_bpath = string_to_atom("bpath");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_ac = string_to_atom("arg_ceg");
	node_ptr a_pred = string_to_atom("pred1");
	node_ptr a_succ = string_to_atom("succ1");
	node_ptr a_len = string_to_atom("length");
	node_ptr a_delarr = string_to_atom("delarr");
	node_ptr a_ce = string_to_atom("ce");
	node_ptr a_push = string_to_atom("push");
	node_ptr a_pop = string_to_atom("pop");
	node_ptr a_top = string_to_atom("top");
	node_ptr a_fsat = string_to_atom("fsat");
	node_ptr a_I = string_to_atom("I");
	node_ptr a_V = string_to_atom("V");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_0 = new_NUMBER_node(0);

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_ac, NIL));

	node_ptr let_bulk1 = new_node(LET, a_bulk, new_node(AND, new_node(NOT, a_ac, NIL), new_node(FUNC, a_pred, cons(a_tsn, cons(num_1, NIL)))));	

	node_ptr loc_newf = new_node(LOCAL, a_newf, a_bulk);
	//node_ptr prt_mcip = new_node(PRINT, cons(new_string_node(QUOTE, "Model checking Invariance Property\n", NIL), NIL), NIL);
	char mcip[] = "Model checking Invariance Property\n";
	node_ptr prt_mcip = new_node(PRINT, cons(new_string_node(QUOTE, mcip, NIL), NIL), NIL);
	node_ptr new_bpath = new_node(RUN, a_new, cons(a_bpath, NIL));
	node_ptr push_b_b = new_node(RUN, a_push, cons(a_bulk, cons(a_bpath, NIL)));
//While(new)	
	node_ptr let_newf1 = new_node(LET, a_newf, new_node(AND, new_node(FUNC, a_pred, cons(a_tsn, cons(a_newf, NIL))), new_node(NOT, a_bulk, NIL)));
	node_ptr push_b_n = new_node(RUN, a_push, cons(a_newf, cons(a_bpath, NIL)));
	node_ptr let_bulk2 = new_node(LET, a_bulk, new_node(OR, a_bulk, a_newf));
	node_ptr ifcond1 = new_node(AND, a_newf, new_node(FUNC, a_I, cons(a_tsn, NIL)));
	node_ptr thenact1 = cons(new_node(LET, a_newf, num_0), NIL);
	node_ptr ifact1 = new_tlv_if(ifcond1, thenact1, NIL);
	
	node_ptr wstmts = cons(let_newf1, cons(push_b_n, cons(let_bulk2, cons(ifact1, NIL))));
	node_ptr wact1 = new_node(WHILE, a_newf, wstmts);

	/*
	char pdbg[] = "\n*** DEBUG ***\n";
	node_ptr prt_dbg = new_node(PRINT, cons(new_string_node(QUOTE, pdbg, NIL), NIL), NIL);

	node_ptr ifcond2 = new_node(AND, a_bulk, new_node(FUNC, a_I, cons(a_tsn, NIL)));
	node_ptr ret0 = new_node(TLV_RETURN, num_0, NIL);
	//node_ptr prt_nv = new_node(PRINT, cons(new_string_node(QUOTE, "\n*** Property is NOT VALID ***\n", NIL), NIL), NIL);
	char pinv[] = "\n*** Property is NOT VALID ***\n";
	node_ptr prt_nv = new_node(PRINT, cons(new_string_node(QUOTE, pinv, NIL), NIL), NIL);
	node_ptr ret1 = new_node(TLV_RETURN, num_1, NIL);
	//node_ptr prt_iv = new_node(PRINT, cons(new_string_node(QUOTE, "\n*** Property is VALID ***\n", NIL), NIL), NIL);
	char piv[] = "\n*** Property is VALID ***\n";
	node_ptr prt_iv = new_node(PRINT, cons(new_string_node(QUOTE, piv, NIL), NIL), NIL);
	node_ptr ifact2 = new_tlv_if(ifcond2, cons(prt_nv, cons(ret0, NIL)), cons(prt_iv, cons(ret1, NIL)));

	//node_ptr prt_bulk = new_node(PRINT, cons(a_bulk, NIL), NIL);
	node_ptr stmts = cons(ifact2, cons(prt_dbg, cons(wact1, cons(push_b_b, cons(new_bpath, cons(prt_mcip, cons(loc_newf, cons(let_bulk1, NIL))))))));
//End while(new)
	func_command(name, param, reverse(stmts), 0);
	*/
//If(bulk &I(tsn))	
	node_ptr ifcond2 = new_node(AND, a_bulk, new_node(FUNC, a_I, cons(a_tsn, NIL)));
	//Then 
	node_ptr prt_nv = new_node(PRINT, cons(new_string_node(QUOTE, "\n*** Property is NOT VALID ***\n", NIL), NIL), NIL);
	node_ptr prt_ce = new_node(PRINT, cons(new_string_node(QUOTE, "\nCounter-Example Follows:\n", NIL), NIL), NIL);
	node_ptr new_ce = new_node(RUN, a_new, cons(a_ce, NIL));
	node_ptr afsat1 = new_node(FUNC, a_fsat, cons(new_node(FUNC, a_V, cons(a_tsn, NIL)), cons(new_node(AND, new_node(FUNC, a_pop, cons(a_bpath, NIL)), new_node(FUNC, a_I, cons(a_tsn, NIL))), NIL))); 
	node_ptr push_ce1 = new_node(RUN, a_push, cons(afsat1, cons(a_ce, NIL)));
		//While (length(ce))
	node_ptr afsat2 = new_node(FUNC, a_fsat, cons(new_node(FUNC, a_V, cons(a_tsn, NIL)), cons(new_node(AND, new_node(FUNC, a_pop, cons(a_bpath, NIL)), new_node(FUNC, a_succ, cons(a_tsn, cons(new_node(FUNC, a_top, cons(a_ce, NIL)), NIL)))), NIL))); 

	node_ptr push_ce2 = new_node(RUN, a_push, cons(afsat2, cons(a_ce, NIL)));
	node_ptr wact2 = new_node(WHILE, new_node(FUNC, a_len, cons(a_bpath, NIL)), cons(push_ce2, NIL));
		//End While (length(ce))
		//For
	node_ptr prt_st = new_node(PRINT, cons(new_node(ARRAY, a_ce, a_i), cons(new_string_node(QUOTE, "=\n", NIL), cons(a_i, cons(new_string_node(QUOTE, "\n--- State no. ", NIL), NIL)))), NIL);
	node_ptr foract = new_tlvfor(a_i, PLUS, num_1, new_node(FUNC, a_len, cons(a_ce, NIL)), cons(prt_st, NIL));
		//End for
	node_ptr ret0 = new_node(TLV_RETURN, num_0, NIL);
	//Else
	node_ptr preact = cons(wact2, cons(push_ce1, cons(new_ce, cons(prt_ce, cons(prt_nv, NIL))))); 
	node_ptr thenact2 = cons(ret0, append(foract, preact));
	node_ptr prt_iv = new_node(PRINT, cons(new_string_node(QUOTE, "\n*** Property is VALID ***\n", NIL), NIL), NIL);
	node_ptr delb = new_node(RUN, a_delarr, cons(a_bpath, NIL));
	node_ptr prt_bulk = new_node(PRINT, a_bulk, NIL);
	node_ptr ret1 = new_node(TLV_RETURN, num_1, NIL);
	node_ptr elseact2 = cons(ret1, cons(delb, cons(prt_iv, NIL)));
	//End Else
	node_ptr ifact2 = new_tlv_if(ifcond2, reverse(thenact2), reverse(elseact2));
//End If

	node_ptr stmts = cons(ifact2, cons(wact1, cons(push_b_b, cons(new_bpath, cons(prt_mcip, cons(loc_newf, cons(let_bulk1, NIL)))))));
	func_command(name, param, reverse(stmts), 0);
}

//Initialize TS
void load_init_TS()
{
	node_ptr name = string_to_atom("init_TS");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_sn = string_to_atom("_sn");
	node_ptr a_n = string_to_atom("n");
	node_ptr a_cur = string_to_atom("cur");
	node_ptr a_top = string_to_atom("top");
	node_ptr num_1 = new_NUMBER_node(1);
	node_ptr num_2 = new_NUMBER_node(2);
	node_ptr num_0 = new_NUMBER_node(0);

	node_ptr param = NIL;
	
	node_ptr ifcond = new_node(NOT, new_node(EXIST, new_node(DOT, a_s, a_n), NIL), NIL);
	node_ptr lsn = new_node(LET, new_node(DOT, a_s, a_n), a_sn);
	node_ptr lst = new_node(LET, new_node(DOT, a_s, a_top), num_0);
	node_ptr conv = new_node(RUN, string_to_atom("convert_old_tlv_to_TS"), cons(num_2, NIL));
	node_ptr thenact = cons(lsn, cons(lst, cons(conv, NIL)));
	node_ptr elseact = cons(new_node(LET, new_node(DOT, a_s, a_top), num_1), NIL);
	node_ptr ifact = new_tlv_if(ifcond, thenact, elseact);

	node_ptr lsc = new_node(LET, new_node(DOT, a_s, a_cur), num_1);

	node_ptr stmts = cons(lsc, cons(ifact, NIL));
	to_command(name, param, reverse(stmts));
}

void load_Invariance()
{
	
	node_ptr name = string_to_atom("Invariance");

	node_ptr a_Cag = string_to_atom("Cag");
	node_ptr a_s = string_to_atom("_s");
	node_ptr a_tsn = string_to_atom("tsn");
	node_ptr a_ai = string_to_atom("arg_inv");
	node_ptr a_res = string_to_atom("result");
	node_ptr a_s_cur = new_node(DOT, a_s, string_to_atom("cur"));

	node_ptr param = cons(new_default_param(a_tsn, a_s_cur), cons(a_ai, NIL));
	node_ptr letact = new_node(LET, a_res, new_node(FUNC, a_Cag, cons(a_tsn, cons(a_ai, NIL))));
	node_ptr stmts = cons(letact, NIL);
	to_command(name, param, reverse(stmts));
}

//Load some built-in commands and functions in the program hash.
void load_build_in_command()
{
	load_new_arr(); 
	load_del_arr();
	load_push_arr();
	load_pop_arr();
	load_top_arr();
	load_length_arr();
	load_new_ts();
	load_set_I();
	load_I();
	load_set_V();
	load_V();
	load_add_disjunct_T();
	load_Ti();
	load_pred1();
	load_succ1();
	load_successors();
	load_predecessors();
	load_reachable();
	load_convert_old_tlv_to_TS();
	load_init_TS();
	load_Cag();
	load_Invariance();
}


// Return the filename portion of a path. The result is a substring, not a copy
static const char *filename_only(const char *path)
{
  char *res = strrchr(path, PATH_SEP);
  return res ? res : path;
}

/**
   Sets the current load path to the absolute path portion of path (which is
   assumed to end in a filename
   If fname is a relative path, current load path is left unchanged

   This overrides the current load path, which is assumed to have been saved
   somewhere.
*/
static void set_current_load_path(const char *fname)
{
  const char *filename = filename_only(fname);
  if (filename > fname) {
    int path_len = strlen(fname) - strlen(filename);
    current_load_path = (char *) malloc(sizeof(char) * (path_len + 1));
    strncpy(current_load_path, fname, path_len);
    current_load_path[path_len] = '\0';
  }
}

static int is_relative_path(const char *path)
{
  return path[0] != PATH_SEP;
}

/// Allocates a new string containing the absolute version of path
static char *make_absolute_path(const char *path, const char *relative_to)
{
  char *res;
  int base_len = strlen(relative_to);
  int sep_missing = (relative_to[base_len - 1] == PATH_SEP) ? 0 : 1;

  res = malloc(sizeof(char) *
               (base_len + sep_missing + strlen(path) + 1));
  strcpy(res, relative_to);
  if (sep_missing)
    res[base_len] = PATH_SEP;
  strcpy(res + base_len + (sep_missing ? 1 : 0), path);
  return res;
}

/* Open a file which contains tlv commands.  If the file is not found
   it is looked for in the TLV directory.
   No parsing is done. The file descriptor is returned. */
FILE *open_file(char *fname, int relative)
{
  static char TLV_PATH_fname[300];
  int free_fname = 0;

  if (is_relative_path(fname) && relative) {
    fname = make_absolute_path(fname, current_load_path);
    free_fname = 1;
  }

  set_current_load_path(fname);

  FILE *in = fopen(fname, "r");

  /* If the file has not been found then try the
     "system" directory */
  if (in == NULL && tlv_path_exists()) {
    add_to_env(TLV_PATH_fname,fname);

    in = fopen(TLV_PATH_fname,"r");
  }
  if (free_fname)
    free(fname);
  return in;
}

/* Parses an input file and stores the result in parse_tree */
int parse_input()
{
  extern node_ptr parse_tree;

  err_node = parse_tree = NIL;
  int status = yyparse();
  if (status)
    printf("Some error while parsing\n");
  return status;
}

/* Read lines of input, using the parser, and execute them. */
void read_and_do(int prompt)
{
  extern node_ptr iparse_tree;

  extern set_prompt();
  extern char do_prompt; /* From grammar.y */

  /* Loop that reads a line and executes it. */
  while (1) {
    err_node = iparse_tree = NIL;

    if (prompt) {
      set_prompt();
      do_prompt = 1;
    }
    else
      do_prompt = 0;

    yyparse(); /* Parse interactive input */

    if (iparse_tree == NIL)
      break; /* End of file */
    else if (iparse_tree != FAILURE_NODE)
      catch_err(run_1_stmt(iparse_tree)); /* Run statement. */
  }
}

typedef struct {
  char *load_path;
  char *input_file;
  int input_fileno;
  FILE *yyin;
  int yylineno;
  node_ptr err_node;
} load_state;

void save_load_state(load_state *state)
{
  extern char *input_file;
  extern int input_fileno;
  extern FILE *yyin;
  extern int yylineno;

  state->input_file = input_file;
  state->input_fileno = input_fileno;
  state->yyin = yyin;
  state->yylineno = yylineno;
  state->err_node = err_node;
  state->load_path = current_load_path;
}

void restore_load_state(load_state *state)
{
  extern char *input_file;
  extern int input_fileno;
  extern FILE *yyin;
  extern int yylineno;

  input_file = state->input_file;
  input_fileno = state->input_fileno;
  yyin = state->yyin;
  yylineno = state->yylineno;
  err_node = state->err_node;

  if (current_load_path != state->load_path) {
    free(current_load_path);
    current_load_path = state->load_path;
  }
}

static int prepare_load(char *fname, int relative)
{
  extern char *input_file;
  extern int input_fileno;
  extern FILE *yyin;
  extern int yylineno;

  FILE *temp;
  if ((temp = open_file(fname, relative)) != NULL) {
    input_file = fname;
    input_fileno = fname2num(fname);
    yyin = temp;
    yylineno = 0;
    err_node = NIL;

    /* Set silent to 1, so that interactive messages will
       not be printed when programs are loaded from files. */
    silent = 1;

    /* Read file. */
    init_input();
    return 1;
  }
  return 0;
}

void finalize_load(load_state *old_load_state)
{
  extern FILE *yyin;

  fclose(yyin);
  restore_load_state(old_load_state);
    /* input_file = old_input_file; */
    /* input_fileno = old_fileno; */
    /* yyin = old_yyin; */
    /* yylineno = old_yylineno; */
    /* err_node = temperr; */
}

/* This procedure loads a file which contains tlv commands
   (this is used for loading the rules file) */
int load_script_file(char *fname, int relative)
{
  /* extern char *input_file; */
  /* extern int input_fileno; */
  /* extern FILE *yyin; */
  /* extern int yylineno; */

  /* FILE *temp; */

  /* char *old_input_file = input_file; */
  /* int old_fileno = input_fileno; */
  /* FILE *old_yyin = yyin; */
  /* int old_yylineno = yylineno; */
  /* node_ptr temperr = err_node; */

  load_state old_load_state;
  save_load_state(&old_load_state);

  if (prepare_load(fname, relative)) {
    read_and_do(0);
    finalize_load(&old_load_state);
    return  1;
  }
  return 0;
}

int parse_file(char *fname, node_ptr *result)
{
  extern node_ptr parse_tree;

  printf("parse_file: Parsing file %s..\n", fname);

  load_state old_load_state;
  save_load_state(&old_load_state);

  if (prepare_load(fname, 0)) {
    parse_input();
    if (parse_tree == NULL)
      printf("parse_file: null!\n");
    *result = parse_tree;
    finalize_load(&old_load_state);
    return  1;
  }
  return 0;
}

/* Load a special rules files. */
void load_rule_file(void)
{
  /* Default name of rule file. */
  static char* rule_filename ="Rules.tlv";

  /* Get name of rule file. */
  char* rule_file_env = getenv("TLV_RULES");

  /* If the environment variable is not defined then default
     to Rules.tlv . */
  if (!rule_file_env)
    rule_file_env = rule_filename;

  /* Load the file. */
  if (load_script_file(rule_file_env, 0)) {
		//Yi Lv
    //fprintf(tlvstdout, "Loaded rules file %s.\n", rule_file_env);
	} else
    fprintf(tlvstderr,"\nThe rule file %s has not been found...\n\n",
            rule_file_env);
}


/* Enter the interactive mode of tlv */
void go_interactive(void)
{
  extern char *input_file;
  extern int input_fileno;
  extern FILE *yyin;
  extern int yylineno;

  fprintf(tlvstdout,"\n      Your wish is my command ... \n\n");

  /*  set_prompt(); */

  input_file = NULL;
  input_fileno = 0;
  yylineno=0;
  yyin = stdin;

  init_input();

  read_and_do(1);
}


/* Initialize script, by initializing hash tables. */
void init_script(void)
{
  progs_hash = new_assoc();
  funcs_hash[0] = new_assoc();
  funcs_hash[1] = new_assoc();
}
